
import os
import tempfile
import pytest

from badwulf.tools import *

def test_to_bytes():
	assert to_bytes(1) == 1
	assert to_bytes(1, "KB") == 1_000
	assert to_bytes(1, "MB") == 1_000_000

def test_format_bytes():
	assert format_bytes(1) == "1 byte"
	assert format_bytes(100) == "100 bytes"
	assert format_bytes(1_000) == "1.0 KB"
	assert format_bytes(1_000_000) == "1.0 MB"

def test_fix_path_err():
	tmpdir = tempfile.gettempdir()
	tmp = os.path.join(tmpdir, "__badwulf_testfile__")
	if os.path.exists(tmp):
		os.remove(tmp)
	with pytest.raises(FileNotFoundError) as err:
		fix_path(tmp)
	assert "path does not exist" in str(err.value)

def test_file_create_remove():
	tmpdir = tempfile.gettempdir()
	tmp = os.path.join(tmpdir, "__badwulf_testfile__")
	if os.path.exists(tmp):
		os.remove(tmp)
	file_create(tmp)
	assert os.path.exists(tmp)
	assert os.path.basename(tmp) in ls(tmpdir)
	file_remove(tmp)
	assert not os.path.exists(tmp)

def test_findport():
	p = findport()
	assert checkport(p) != 0

@pytest.fixture
def string_list():
	return [
		"I am the Bad Wolf.",
		"I create myself.",
		"I take the words, I scatter them in time and space.",
		"A message to lead myself here."]

def test_grep1(string_list):
	q = grep1("bad wolf", string_list[0])
	assert q is not None
	assert q.span() == (9, 17)

def test_grep1_context(string_list):
	q1 = grep1("bad wolf", string_list[0], context_width=8)
	assert q1 == "Bad Wolf"
	q2 = grep1("bad wolf", string_list[0], context_width=20)
	assert q2 == string_list[0]

def test_grepl(string_list):
	qs = grepl("bad wolf", string_list)
	assert qs == [True, False, False, False]




from badwulf import rssh

def test_rssh_without_gateway():
	con = rssh("bad-wolf", "vortex")
	assert con.username == "bad-wolf"
	assert con.destination == "vortex"
	assert con.hostname == "vortex"
	assert con.server is None
	assert con.server_username == "bad-wolf"
	assert not con.isopen()

def test_rssh_with_gateway():
	con = rssh("bad-wolf", "vortex",
		server="login.dimension.time",
		server_username="root",
		autoconnect=False)
	assert con.username == "bad-wolf"
	assert con.destination == "vortex"
	assert con.hostname == "localhost"
	assert con.server == "login.dimension.time"
	assert con.server_username == "root"
	assert not con.isopen()




# Tools and utility functions

import os
import re
import platform
import socket
import random
from importlib.metadata import version

def badwulf_version():
	"""
	Get badwulf package version
	"""
	return version("badwulf")

def badwulf_attribution():
	"""
	Get badwulf package attribution
	"""
	str1 = "powered by badwulf |"
	str2 = "copyright (c) 2024 Kylie A. Bemis"
	return str1 + " " + str2

def is_known_host(nodes):
	"""
	Check if the program is running on a known host
	:param nodes: A list of hostnames
	:returns: True if running on a known host, False otherwise
	"""
	host = platform.node().replace(".local", "")
	nodes = [nodename.casefold() for nodename in nodes]
	return host.casefold() in nodes

def to_bytes(x, units = "bytes"):
	"""
	Convert a size to bytes
	:param x: A positive number
	:param units: The units for x (KB, MB, GB, etc.)
	:returns: The number of bytes
	"""
	if units in ("bytes", "B"):
		pass
	elif units == "KB":
		x *= 1000
	elif units == "MB":
		x *= 1000 ** 2
	elif units == "GB":
		x *= 1000 ** 3
	elif units == "TB":
		x *= 1000 ** 4
	elif units == "PB":
		x *= 1000 ** 5
	else:
		raise ValueError(f"invalid units: {units}")
	return x

def format_bytes(x, units = "auto"):
	"""
	Format bytes
	:param x: The number of bytes
	:param units: The units (B, KB, MB, etc.)
	:returns: A string
	"""
	if units == "auto":
		if x >= 1000 ** 5:
			units = "PB"
		elif x >= 1000 ** 4:
			units = "TB"
		elif x >= 1000 ** 3:
			units = "GB"
		elif x >= 1000 ** 2:
			units = "MB"
		elif x >= 1000:
			units = "KB"
		else:
			units = "bytes"
	if units in ("bytes", "B"):
		if x == 1 and units == "bytes":
			units = "byte"
		x = int(x)
	else:
		if units == "KB":
			x /= 1000
		elif units == "MB":
			x /= 1000 ** 2
		elif units == "GB":
			x /= 1000 ** 3
		elif units == "TB":
			x /= 1000 ** 4
		elif units == "PB":
			x /= 1000 ** 5
		else:
			raise ValueError(f"invalid units: {units}")
		x = round(x, ndigits=2)
		x = float(x)
	return f"{x} {units}"

def print_bytes(x, units = "auto"):
	"""
	Print bytes
	:param x: The number of bytes
	:param units: The units (B, KB, MB, etc.)
	"""
	print(format_bytes(x, units))

def format_datasets(iterable, names_only = False, header = True):
	"""
	Format datasets
	:param iterable: An iterable of datasets
	:param names_only: Print names only?
	:param header: Print number of datasets?
	:return: A formatted string
	"""
	if names_only:
		sl = [f"['{dataset.name}']" 
			for dataset 
			in iterable]
	else:
		sl = [f"['{dataset.name}']\n{dataset}" 
			for dataset 
			in iterable]
	if header:
		sl = [f"#### {len(sl)} datasets ####\n"] + sl
	return "\n".join(sl)

def print_datasets(iterable, names_only = False, header = True):
	"""
	Print datasets
	:param iterable: An iterable of datasets
	:param names_only: Print names only?
	:param header: Print number of datasets?
	"""
	print(format_datasets(iterable, names_only, header))

def askYesNo(msg = "Continue? (yes/no): "):
	"""
	Ask a user to confirm yes or no
	:param msg: The message to print
	:returns: True if yes, False if no
	"""
	while True:
		confirm = input(msg).casefold()
		if confirm in ("y", "yes"):
			return True
		elif confirm in ("n", "no"):
			return False
		else:
			print("Invalid input. Please enter yes/no.")

def fix_path(path, must_exist = True):
	"""
	Normalize and expand paths
	:param path: The path to normalize
	:param must_exist: Must the path exist?
	:returns: The normalized path
	"""
	if "~" in path:
		path = os.path.expanduser(path)
	path = os.path.realpath(path)
	if must_exist and not os.path.exists(path):
		raise FileNotFoundError(f"path does not exist: '{path}'")
	return path

def file_create(path):
	"""
	Create a file
	:param path: The file to create
	"""
	path = fix_path(path, must_exist=False)
	with open(path, "a"):
		os.utime(path, None)

def file_remove(path):
	"""
	Delete a file
	:param path: The file to delete
	"""
	path = fix_path(path, must_exist=False)
	if os.path.exists(path):
		os.remove(path)

def ls(path = ".", all_names = False):
	"""
	List files in a directory
	:param path: The directory
	:param all_names: Should hidden files be included?
	:returns: A list of file names
	"""
	path = fix_path(path)
	if not os.path.isdir(path):
		raise NotADirectoryError(f"path must be a directory: {path}")
	if all_names:
		return [f 
			for f 
			in os.listdir(path)]
	else:
		return [f 
			for f 
			in os.listdir(path)
			if not f.startswith(".")]

def dirsize(path, all_names = False):
	"""
	Get size of a directory
	:param path: The directory
	:param all_names: Should hidden files be included?
	:returns: The size of the directory in bytes
	"""
	size = 0
	files = ls(path, all_names=all_names)
	for file in files:
		if file in (".", ".."):
			continue
		file = os.path.join(path, file)
		if os.path.isdir(file):
			size += dirsize(file, all_names=all_names)
		else:
			size += os.path.getsize(file)
	return size

def dirfiles(path, pattern, recursive = False, all_names = False):
	"""
	Get files in a directory matching a pattern
	:param path: The directory
	:param pattern: The pattern
	:param all_names: Should hidden files be included?
	:returns: The size of the directory in bytes
	"""
	matches = []
	files = ls(path, all_names=all_names)
	for file in files:
		if file in (".", ".."):
			continue
		file = os.path.join(path, file)
		if os.path.isdir(file) and recursive:
			matches.extend(dirfiles(file, pattern, all_names=all_names))
		elif grep1(pattern, file) is not None:
			matches.append(file)
	return matches

def checkport(port):
	"""
	Check if a port is open (i.e., if it is in use)
	:param port: The port to check
	:returns: 0 if open, an error code otherwise
	"""
	sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	result = sock.connect_ex(("localhost", port))
	sock.close()
	return result

def findport(attempts = 10):
	"""
	Find an available port (for SSH forwarding)
	:param attempts: How many random ports to attempt
	:returns: The port number
	"""
	for i in range(attempts):
		port = random.randint(1024, 65535)
		if checkport(port) != 0:
			return port
	raise IOError("couldn't find an available port")

def grep1(pattern, x, ignore_case = True, context_width = None):
	"""
	Search for a pattern in a string
	:param pattern: The pattern to find
	:param x: A string
	:param ignore_case: Should case be ignored?
	:param context_width: Width of a context window to return
	:returns: A Match or None
	"""
	if x is None:
		return None
	if ignore_case:
		match = re.search(pattern, x, flags=re.IGNORECASE)
	else:
		match = re.search(pattern, x)
	if match is None or context_width is None:
		return match
	else:
		start = match.start()
		stop = match.end()
		margin = ((context_width - (stop - start)) // 2)
		if context_width > len(x):
			return x
		if context_width < stop - start or margin < 4:
			return x[start:stop]
		pre, post = "", ""
		if start > margin:
			start = max(0, start - margin)
			pre = "..."
		if len(x) - stop > margin:
			stop = min(len(x), stop + margin)
			post = "..."
		return pre + x[start:stop] + post

def grep(pattern, x, ignore_case = True):
	"""
	Search for a pattern in an iterable
	:param pattern: The pattern to find
	:param x: An iterable
	:param ignore_case: Should case be ignored?
	:returns: A list of matches
	"""
	if x is None:
		return []
	else:
		return [grep1(pattern, xi, ignore_case=ignore_case) 
			for xi 
			in x]

def grepl(pattern, x, ignore_case = True):
	"""
	Search for a pattern in an iterable
	:param pattern: The pattern to find
	:param x: An iterable
	:param ignore_case: Should case be ignored?
	:returns: A list of bools
	"""
	if x is None:
		return []
	else:
		return [match is not None 
			for match 
			in grep(pattern, x, ignore_case=ignore_case)]





import sys
import platform
import subprocess
import argparse
import datetime
from time import sleep

from ..rssh import rssh
from ..tools import badwulf_attribution
from ..tools import is_known_host
from ..tools import findport

class clmanager:
	"""
	Command line utility for Beowulf clusters
	"""
	
	def __init__(self,
		name,
		nodes,
		version,
		date,
		description,
		readme = None,
		program = None,
		username = None,
		server = None,
		server_username = False,
		port = None):
		"""
		Initialize a cluster CLI utility program
		:param name: The name of the cluster/server
		:param nodes: A list of nodenames or dict in the form {alias: nodename}
		:param version: The version of the program
		:param date: The date of the program's last revision
		:param description: A description of the program
		:param readme: The file path of a README.md file
		:param program: The name of the program (defaults to name)
		:param username: Your username on the cluster
		:param server: The gateway server hostname (optional)
		:param server_username: Your username on the gateway server (optional)
		:param port: The local port for gateway server SSH forwarding
		"""
		self.name = name
		self.nodes = nodes
		self.version = version
		if isinstance(date, datetime.date):
			self.date = date
		else:
			self.date = datetime.date.fromisoformat(date)
		self.description = description
		self.readme = readme
		if program is None:
			self.program = name.casefold()
		else:
			self.program = program
		self.username = username
		self.server = server
		self.server_username = server_username
		self.port = port
		self._parser = None
		self._args = None

	def _add_cluster_args(self, parser):
		"""
		Add cluster parameters to a parser.
		:param parser: The parser to update
		"""
		if isinstance(self.nodes, dict):
			for alias, nodename in self.nodes.items():
				parser.add_argument(f"-{alias}", action="append_const",
					help=nodename, dest="nodes", const=nodename)
		parser.add_argument("-n", "--node", action="append",
			help=f"{self.name} node", dest="nodes",
			metavar="NODE")
		parser.add_argument("-p", "--port", action="store",
			help="port forwarding", default=self.port)
		parser.add_argument("-u", "--user", action="store",
			help=f"{self.name} user", default=self.username)
		parser.add_argument("-L", "--login", action="store",
			help="gateway server user", default=self.server_username)
		parser.add_argument("-S", "--server", action="store",
			help="gateway server host", default=self.server)
	
	def _add_subcommand_run(self, subparsers):
		"""
		Add 'run' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("run", 
			help=f"run command (e.g., shell) on a {self.name} node")
		self._add_cluster_args(cmd)
		cmd.add_argument("remote_command", action="store",
			help="command to execute on a Magi node", nargs=argparse.OPTIONAL,
			metavar="command")
		cmd.add_argument("remote_args", action="store",
			help="command arguments", nargs=argparse.REMAINDER,
			metavar="...")
	
	def _add_subcommand_copy_id(self, subparsers):
		"""
		Add 'copy-id' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("copy-id", 
			help=f"copy ssh keys to a {self.name} node")
		self._add_cluster_args(cmd)
		cmd.add_argument("identity_file", action="store",
			help="ssh key identity file")
	
	def _add_subcommand_upload(self, subparsers):
		"""
		Add a 'upload' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("upload", 
			help=f"upload file(s) to {self.name}")
		self._add_cluster_args(cmd)
		cmd.add_argument("src", action="store",
			help="source file/directory")
		cmd.add_argument("dest", action="store",
			help="destination file/directory")
		cmd.add_argument("--ask", action="store_true",
			help="ask to confirm before uploading files?")
		cmd.add_argument("--dry-run", action="store_true",
			help="show what would happen without doing it?")
	
	def _add_subcommand_download(self, subparsers):
		"""
		Add a 'download' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("download", 
			help=f"download file(s) from {self.name}")
		self._add_cluster_args(cmd)
		cmd.add_argument("src", action="store",
			help="source file/directory")
		cmd.add_argument("dest", action="store",
			help="destination file/directory")
		cmd.add_argument("--ask", action="store_true",
			help="ask to confirm before downloading files?")
		cmd.add_argument("--dry-run", action="store_true",
			help="show what would happen without doing it?")
	
	def _add_subcommand_readme(self, subparsers):
		"""
		Add 'readme' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("readme", 
			help="display readme")
		cmd.add_argument("-p", "--pager", action="store",
			help="program to display readme (default 'glow')")
		cmd.add_argument("-w", "--width", action="store",
			help="word-wrap readme at width (default 70)", default=70)

	def _init_parser(self):
		"""
		Initialize the argument parser
		"""
		parser = argparse.ArgumentParser(self.program,
			description=self.description)
		parser.add_argument("-v", "--version", action="store_true",
			help="display version")
		subparsers = parser.add_subparsers(dest="cmd")
		self._add_subcommand_run(subparsers)
		self._add_subcommand_copy_id(subparsers)
		self._add_subcommand_upload(subparsers)
		self._add_subcommand_download(subparsers)
		if self.readme is not None:
			self._add_subcommand_readme(subparsers)
		self._parser = parser
	
	def is_node(self):
		"""
		Check if the program is running on a cluster node
		:returns: True if running the a cluster node, False otherwise
		"""
		if isinstance(self.nodes, dict):
			nodes = self.nodes.values()
		else:
			nodes = self.nodes
		return is_known_host(nodes)
	
	def resolve_node(self, nodes):
		"""
		Get single valid nodename from a list of nodes
		:param nodes: A list of nodenames
		"""
		host = platform.node().replace(".local", "")
		if nodes is None or len(nodes) != 1:
			sys.exit(f"{self.program}: error: must specify exactly _one_ {self.name} node")
		node = nodes[0]
		if self.is_node():
			if host == node.casefold():
				node = "localhost"
			else:
				node += ".local"
		return node
	
	def open_ssh(self,
		node,
		username = None,
		server = None,
		server_username = None,
		port = None):
		"""
		Open SSH connection to a cluster node
		:param node: The target cluster node
		:param username: Your username on the cluster
		:param server: The gateway server hostname (optional)
		:param server_username: Your username on the gateway server (optional)
		:param port: Port used for gateway forwarding
		:returns: An open rssh instance
		"""
		if username is None:
			username = self.username
		if server is None:
			server = self.server
		if server_username is None:
			server_username = self.server_username
		if port is None:
			port = findport()
		# connect and return the session
		session = rssh(username, node,
			server=server,
			server_username=server_username,
			port=port,
			autoconnect=True)
		return session
	
	def parse_args(self):
		"""
		Parse command line arguments
		"""
		if self._parser is None:
			self._init_parser()
		self._args = self._parser.parse_args()
	
	def main(self):
		"""
		Run the program
		"""
		if self._args is None:
			self.parse_args()
		args = self._args
		# version
		if args.version:
			description = self.description.splitlines()[0]
			print(f"{description} version {self.version} (revised {self.date})")
			print(badwulf_attribution())
			sys.exit()
		# open ssh for server commands
		if args.cmd in ("run", "copy-id", "upload", "download"):
			con = self.open_ssh(self.resolve_node(args.nodes),
				username=args.user,
				server=args.server,
				server_username=args.login, 
				port=args.port)
			sleep(1) # allow time to connect
		# help
		if args.cmd is None:
			self._parser.print_help()
		# run
		elif args.cmd == "run":
			if args.remote_command is None:
				con.ssh()
			else:
				print(f"connecting as {con.username}@{con.destination}")
				dest = f"{con.username}@{con.hostname}"
				if con.server is None:
					cmd = ["ssh", dest]
				else:
					cmd = ["ssh", "-o", "NoHostAuthenticationForLocalhost=yes"]
					cmd += ["-p", str(con.port), dest]
				cmd.append(args.remote_command)
				cmd.extend(args.remote_args)
				subprocess.run(cmd)
		# copy-id
		elif args.cmd == "copy-id":
			con.copy_id(args.identity_file)
		# upload
		elif args.cmd == "upload":
			con.upload(args.src, args.dest,
				dry_run=args.dry_run, ask=args.ask)
		# download
		elif args.cmd == "download":
			con.download(args.src, args.dest,
				dry_run=args.dry_run, ask=args.ask)
		# readme
		elif args.cmd == "readme":
			if args.pager is None:
				cmd = ["glow", "-p", "-w", str(args.width)]
			else:
				cmd = [args.pager]
			cmd += [self.readme]
			subprocess.run(cmd)
		sys.exit()




import os
import sys
import subprocess
import argparse
import datetime

from ..expdb import expdb
from ..tools import badwulf_attribution
from ..tools import print_datasets
from ..tools import format_bytes
from ..tools import findport

class dbmanager:
	"""
	Command line utility for scientific dataset management
	"""
	
	def __init__(self,
		name,
		dbpath,
		dbname,
		version,
		date,
		description,
		readme = None,
		program = None,
		metaname = True,
		username = None,
		remote_dbhost = None,
		remote_dbpath = None,
		server = None,
		server_username = False,
		port = None):
		"""
		Initialize a cluster CLI utility program
		:param name: The name of the cluster/server
		:param dbpath: The local database path
		:param dbname: The database name
		:param version: The version of the program
		:param date: The date of the program's last revision
		:param description: A description of the program
		:param readme: The file path of a README.md file
		:param program: The name of the program (defaults to name)
		:param metaname: Subdirectory containing "manifest.toml"" (optional)
		:param username: Your username on the cluster
		:param remote_dbhost: The remote database host
		:param remote_dbpath: The remote database path
		:param server: The gateway server hostname (optional)
		:param server_username: Your username on the gateway server (optional)
		:param port: The local port for gateway server SSH forwarding
		"""
		self.name = name
		self.dbpath = dbpath
		self.dbname = dbname
		self.version = version
		if isinstance(date, datetime.date):
			self.date = date
		else:
			self.date = datetime.date.fromisoformat(date)
		self.description = description
		self.readme = readme
		if program is None:
			self.program = name.casefold()
		else:
			self.program = program
		self.metaname = metaname
		self.username = username
		self.remote_dbhost = remote_dbhost
		self.remote_dbpath = remote_dbpath
		self.server = server
		self.server_username = server_username
		if port is None:
			self.port = findport()
		else:
			self.port = port
		self._parser = None
		self._args = None

	def _add_server_args(self, parser):
		"""
		Add server parameters to a parser.
		:param parser: The parser to update
		"""
		parser.add_argument("-p", "--port", action="store",
			help="port forwarding", default=self.port)
		parser.add_argument("-u", "--user", action="store",
			help=f"{self.name} user", default=self.username)
		parser.add_argument("-L", "--login", action="store",
			help="gateway server user", default=self.server_username)
		parser.add_argument("-S", "--server", action="store",
			help="gateway server host", default=self.server)
		parser.add_argument("--remote-host", action="store",
			help="remote database host", default=self.remote_dbhost)
		parser.add_argument("--remote-path", action="store",
			help="remote database path", default=self.remote_dbpath)
	
	def _add_subcommand_ls(self, subparsers):
		"""
		Add 'ls' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("ls", 
			help="list all datasets")
		cmd.add_argument("-l", "--long", action="store_true",
			help="show extended details")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
	
	def _add_subcommand_ls_cache(self, subparsers):
		"""
		Add 'ls-cache' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("ls-cache", 
			help="list cached datasets")
		cmd.add_argument("-l", "--long", action="store_true",
			help="show extended details")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
		cmd.add_argument("--sort", action="store",
			help="sort by file attribute (atime, mtime, size)")
		cmd.add_argument("--reverse", action="store",
			help="reverse by file attribute (atime, mtime, size)")
	
	def _add_subcommand_search(self, subparsers):
		"""
		Add 'search' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("search", 
			help="search all datasets")
		cmd.add_argument("pattern", action="store",
			help="search pattern (regex allowed)")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
	
	def _add_subcommand_search_cache(self, subparsers):
		"""
		Add 'search-cache' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("search-cache", 
			help="search cached datasets")
		cmd.add_argument("pattern", action="store",
			help="search pattern (regex allowed)")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
		
	def _add_subcommand_prune_cache(self, subparsers):
		"""
		Add 'prune-cache' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("prune-cache", 
			help="remove cached datasets")
		cmd.add_argument("limit", action="store",
			help="maximum cache size (10, 100, 1000, etc.)", type=float)
		cmd.add_argument("units", action="store",
			help="cache size units (MB, GB, TB, etc.)")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
		cmd.add_argument("-L", "--lru", action="store_const",
			help="prune least recently used (default)", dest="strategy", const="lru")
		cmd.add_argument("-M", "--mru", action="store_const",
			help="prune most recently used", dest="strategy", const="mru")
		cmd.add_argument("-B", "--big", action="store_const",
			help="prune largest files", dest="strategy", const="big")
		cmd.add_argument("-S", "--small", action="store_const",
			help="prune smallest files", dest="strategy", const="small")
		cmd.add_argument("--ask", action="store_true",
			help="ask to confirm before deleting?")
		cmd.add_argument("--dry-run", action="store_true",
			help="show what would happen without doing it?")
	
	def _add_subcommand_describe(self, subparsers):
		"""
		Add 'describe' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("describe", 
			help="describe a dataset")
		cmd.add_argument("id", action="store",
			help="the identifier of the dataset to describe")
	
	def _add_subcommand_sync(self, subparsers):
		"""
		Add 'sync' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("sync", 
			help="sync a dataset to local cache")
		cmd.add_argument("id", action="store",
			help="the identifier of the dataset to sync")
		self._add_server_args(cmd)
		cmd.add_argument("-f", "--force", action="store_true",
			help="force re-sync if already cached")
		cmd.add_argument("--ask", action="store_true",
			help="ask to confirm before syncing?")
		cmd.add_argument("--dry-run", action="store_true",
			help="show what would happen without doing it?")
	
	def _add_subcommand_submit(self, subparsers):
		"""
		Add 'submit' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("submit", 
			help="submit a dataset")
		cmd.add_argument("path", action="store",
			help="the path to the dataset directory")
		self._add_server_args(cmd)
		cmd.add_argument("-f", "--force", action="store_true",
			help="force re-submission if already tracked")
		cmd.add_argument("--ask", action="store_true",
			help="ask to confirm before submitting?")
		cmd.add_argument("--dry-run", action="store_true",
			help="show what would happen without doing it?")
	
	def _add_subcommand_status(self, subparsers):
		"""
		Add 'status' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("status", 
			help="report status of cache against manifest")
		cmd.add_argument("-l", "--long", action="store_true",
			help="show extended details")
		cmd.add_argument("-s", "--scope", action="store",
			help="filter by scope")
		cmd.add_argument("-g", "--group", action="store",
			help="filter by group")
	
	def _add_subcommand_readme(self, subparsers):
		"""
		Add 'readme' subcommand to subparsers.
		:param subparsers: The subparsers to update
		"""
		cmd = subparsers.add_parser("readme", 
			help="display readme")
		cmd.add_argument("-p", "--pager", action="store",
			help="program to display readme (default 'glow')")
		cmd.add_argument("-w", "--width", action="store",
			help="word-wrap readme at width (default 70)", default=70)

	def _init_parser(self):
		"""
		Initialize the argument parser
		"""
		parser = argparse.ArgumentParser(self.program,
			description=self.description)
		parser.add_argument("-v", "--version", action="store_true",
			help="display version")
		subparsers = parser.add_subparsers(dest="cmd")
		self._add_subcommand_ls(subparsers)
		self._add_subcommand_ls_cache(subparsers)
		self._add_subcommand_search(subparsers)
		self._add_subcommand_search_cache(subparsers)
		self._add_subcommand_prune_cache(subparsers)
		self._add_subcommand_describe(subparsers)
		self._add_subcommand_sync(subparsers)
		self._add_subcommand_submit(subparsers)
		self._add_subcommand_status(subparsers)
		if self.readme is not None:
			self._add_subcommand_readme(subparsers)
		self._parser = parser
	
	def parse_args(self):
		"""
		Parse command line arguments
		"""
		if self._parser is None:
			self._init_parser()
		self._args = self._parser.parse_args()
	
	def open_db(self):
		"""
		Open connection to the database
		:param dbpath: The local database path
		"""
		# check for valid database path
		if not os.path.isdir(self.dbpath):
			raise NotADirectoryError(f"database does not exist: '{self.dbpath}'")
		# connect and return database
		db = expdb(self.username, self.dbpath, self.dbname,
			metaname=self.metaname,
			remote_dbhost=self.remote_dbhost,
			remote_dbpath=self.remote_dbpath,
			server=self.server,
			server_username=self.server_username,
			port=self.port,
			verbose=False,
			autoconnect=False)
		return db
	
	def main(self):
		"""
		Run the program
		"""
		if self._args is None:
			self.parse_args()
		args = self._args
		# version
		if args.version:
			description = self.description.splitlines()[0]
			print(f"{description} version {self.version} (revised {self.date})")
			print(badwulf_attribution())
			sys.exit()
		# help
		if args.cmd is None:
			self._parser.print_help()
		else:
			db = self.open_db()
		# ls
		if args.cmd == "ls":
			datasets = db.ls(
				scope=args.scope,
				group=args.group,
				details=args.long)
			if args.long:
				print_datasets(datasets)
			else:
				for name in datasets:
					print(f"['{name}']")
		# ls-cache
		elif args.cmd == "ls-cache":
			sort = args.sort is not None or args.reverse is not None
			datasets = db.ls_cache(
				scope=args.scope,
				group=args.group,
				details=args.long or sort)
			if sort:
				if args.reverse is not None:
					args.sort = args.reverse
					args.reverse = True
				sortby = args.sort.casefold()
				if sortby == "size".casefold():
					datasets.sort(key=lambda x: x.size)
				elif sortby == "atime".casefold():
					datasets.sort(key=lambda x: x.atime)
				elif sortby == "mtime".casefold():
					datasets.sort(key=lambda x: x.mtime)
				else:
					sys.exit(f"msi ls-cache: error: can't sort by attribute: '{args.sort}'")
				if args.reverse:
					datasets.reverse()
			if sort or args.long:
				print_datasets(datasets)
				sizes = [x.size for x in datasets]
				print(f"~= {format_bytes(sum(sizes))} total")
			else:
				for name in datasets:
					print(f"['{name}']")
		# search
		elif args.cmd == "search":
			hits = db.search(
				pattern=args.pattern,
				scope=args.scope,
				group=args.group)
			print_datasets(hits)
		# search-cache
		elif args.cmd == "search-cache":
			hits = db.search_cache(
				pattern=args.pattern,
				scope=args.scope,
				group=args.group)
			print_datasets(hits)
		# prune-cache
		elif args.cmd == "prune-cache":
			if args.strategy is None:
				args.strategy = "lru"
			db.prune_cache(
				limit=args.limit,
				units=args.units,
				scope=args.scope,
				group=args.group,
				strategy=args.strategy,
				dry_run=args.dry_run, ask=args.ask)
		# describe
		elif args.cmd == "describe":
			dataset = db.get(args.id)
			if dataset is None:
				sys.exit(f"msi describe: error: no such dataset: '{args.id}'")
			else:
				print(dataset.describe())
				if args.id in db.cache:
					print(db.cache.get(args.id).describe(dataset.formats))
		# sync
		elif args.cmd == "sync":
			if args.id in db.cache and not args.force:
				sys.exit("msi sync: dataset is already cached; use --force to re-sync")
			db.username = args.user
			db.remote_dbhost = args.remote_host
			db.remote_dbpath = args.remote_path
			db.server = args.server
			db.server_username = args.login
			db.port = args.port
			db.sync(args.id,
				force=args.force,
				dry_run=args.dry_run, ask=args.ask)
		# submit
		elif args.cmd == "submit":
			name = os.path.basename(args.path)
			if name in db.manifest and not args.force:
				sys.exit("msi submit: dataset is already tracked; use --force to re-submit")
			db.username = args.user
			db.remote_dbhost = args.remote_host
			db.remote_dbpath = args.remote_path
			db.server = args.server
			db.server_username = args.login
			db.port = args.port
			db.submit(args.path,
				force=args.force,
				dry_run=args.dry_run, ask=args.ask)
		# status
		elif args.cmd == "status":
			synced, remoteonly, localonly = db.status(
				scope=args.scope,
				group=args.group,
				details=args.long)
			msg_synced = "\n~~~~ synced:"
			msg_remoteonly = "\n>>>> tracked but not cached:"
			msg_localonly = "\n<<<< cached but not tracked:"
			if args.long:
				print(msg_synced)
				print_datasets(synced)
				print(msg_remoteonly)
				print_datasets(remoteonly)
				print(msg_localonly)
				print_datasets(localonly)
			else:
				print(msg_synced)
				for name in synced:
					print(f"['{name}']")
				print(msg_remoteonly)
				for name in remoteonly:
					print(f"['{name}']")
				print(msg_localonly)
				for name in localonly:
					print(f"['{name}']")
		# readme
		elif args.cmd == "readme":
			if args.pager is None:
				cmd = ["glow", "-p", "-w", str(args.width)]
			else:
				cmd = [args.pager]
			cmd += [self.readme]
			subprocess.run(cmd)
		sys.exit()




import os
import pytest

from badwulf import expdb

@pytest.fixture
def database():
	try:
		dbpath = os.path.dirname(__file__)
		dbpath = os.path.join(dbpath, "testdb")
	except:
		dbpath = os.path.join("tests", "testdb")
	return expdb("bad-wolf", dbpath, None, autoconnect=False)

def test_db_open(database):
	with database as db:
		assert db.isopen()

def test_db_get(database):
	with database as db:
		assert db.get("Test-data-1") == db["Test-data-1"]

def test_db_ls(database):
	with database as db:
		assert "Test-data-1" in db.ls()

def test_db_ls_cache(database):
	with database as db:
		assert len(db.ls_cache()) == 0

def test_db_search(database):
	with database as db:
		assert len(db.search("bad wolf")) == 1
		assert len(db.search("myself")) == 2
		assert len(db.search("secret")) == 0

def test_db_search_cache(database):
	with database as db:
		assert len(db.search_cache("bad wolf")) == 0
		assert len(db.search_cache("myself")) == 0
		assert len(db.search_cache("secret")) == 0

def test_db_describe(database):
	with database as db:
		assert db["Test-data-1"].describe() is not None
		assert db["Test-data-2"].describe() is not None
		assert db["Test-data-3"].describe() is not None
		assert db["Test-data-4"].describe() is not None

def test_db_status(database):
	with database as db:
		synced, remoteonly, localonly = db.status()
		assert len(synced) == 0
		assert len(remoteonly) == 4
		assert len(localonly) == 0




# Restricted SSH manager

import subprocess

from .tools import fix_path
from .tools import askYesNo

class rssh:
	"""
	Restricted SSH manager for rsync
	"""
	
	def __init__(self, username, destination,
		server = None, server_username = True,
		port = 8080, destination_port = 22,
		autoconnect = True):
		"""
		Initialize an rssh instance
		:param username: Your username on destination machine
		:param destination: The destination machine hostname
		:param server: The gateway server hostname (optional)
		:param server_username: Your username on the gateway server (optional)
		:param port: The local port for gateway server SSH forwarding
		:param destination_port: The destination port
		:param autoconnect: Connect on initialization?
		"""
		if server_username in (True, False):
			if server_username:
				server_username = username
			else:
				server_username = ""
		self.username = username
		self.destination = destination
		self.server = server
		self.server_username = server_username
		self.port = port
		self.destination_port = destination_port
		self.process = None
		if autoconnect:
			self.open()
	
	def __str__(self):
		"""
		Return str(self)
		"""
		dest = f"{self.username}@{self.destination}"
		if self.isopen():
			server = f"(forwarding to {self.server} over {self.port})"
			return "ssh: " + dest + " " + server
		else:
			return "ssh: " + dest
	
	def __repr__(self):
		"""
		Return repr(self)
		"""
		user = f"username='{self.username}'"
		dest = f"destination='{self.destination}'"
		if self.isopen():
			server = f"server='{self.server}'"
			server_username = f"server_username='{self.server_username}'"
			port = f"port={self.port}"
			return f"rssh({user}, {dest}, {server}, {server_username}, {port})"
		else:
			return f"rssh({user}, {dest})"
	
	def __enter__(self):
		"""
		Enter context manager
		"""
		self.open()
		return self
	
	def __exit__(self, exc_type, exc_value, traceback):
		"""
		Exit context manager
		"""
		self.close()
	
	def __del__(self):
		"""
		Delete self
		"""
		self.close()
	
	@property
	def hostname(self):
		"""
		Get the hostname used for the SSH connection
		"""
		if self.server is None:
			return self.destination
		else:
			return "localhost"
	
	def isopen(self):
		"""
		Check if the gateway server connection is open
		"""
		return self.process is not None

	def open(self):
		"""
		Open the connection to the gateway server
		"""
		if self.server is None or self.isopen():
			return
		print(f"opening connection to {self.server}")
		if not isinstance(self.server_username, str):
			msg = "Please enter your username: "
			self.server_username = input(msg)
		if len(self.server_username) == 0:
			gateway = self.server
		else:
			gateway = f"{self.server_username}@{self.server}"
		target = f"{self.port}:{self.destination}:{self.destination_port}"
		cmd = ["ssh", "-NL", target, gateway]
		try:
			print(f"connecting as {gateway}")
			self.process = subprocess.Popen(cmd)
			print(f"forwarding to {self.destination} on port {self.port}")
		except Exception:
			self.process = None
			print("failed to open connection")
	
	def ls(self, file = None, all_names = False, details = False):
		"""
		List files on the destination machine
		:param file: A file or directory or list of them
		:param all_names: Should hidden files be included?
		:param details: Show file metadata details?
		"""
		host = f"{self.username}@{self.hostname}"
		if self.server is None:
			cmd = ["ssh", host]
		else:
			cmd = ["ssh", "-o", "NoHostAuthenticationForLocalhost=yes"]
			cmd += ["-p", str(self.port), host]
		cmd += ["ls"]
		if all_names:
			cmd += ["-a"]
		if details:
			cmd += ["-l"]
		if file is not None:
			if isinstance(file, str):
				cmd += [file]
			else:
				cmd += file
		print(f"connecting as {self.username}@{self.destination}")
		return subprocess.run(cmd)
	
	def copy_id(self, id_file, ask = False):
		"""
		Copy local SSH keys to the destination machine
		:param id_file: The identity file (ending in .pub)
		:param ask: Confirm before copying?
		"""
		truehost = f"{self.username}@{self.hostname}"
		showhost = f"{self.username}@{self.destination}"
		id_file = fix_path(id_file, must_exist=True)
		print(f"key will be uploaded from: '{id_file}'")
		print(f"key will be uploaded to: '{showhost}'")
		if ask and not askYesNo():
			return
		print(f"copying key as {showhost}")
		cmd = ["ssh-copy-id", "-i", id_file]
		if self.server is None:
			cmd += [truehost]
		else:
			cmd += ["-o", "NoHostAuthenticationForLocalhost=yes"]
			cmd += ["-p", str(self.port)]
			cmd += [truehost]
		return subprocess.run(cmd)
	
	def download(self, src, dest, dry_run = False, ask = False):
		"""
		Download file(s) to local storage using rsync
		:param src: The source path on the destination machine
		:param dest: The destination path on the local machine
		:param dry_run: Show what would be done without doing it?
		:param ask: Confirm before downloading?
		"""
		truesrc = f"{self.username}@{self.hostname}:{src}"
		showsrc = f"{self.username}@{self.destination}:{src}"
		has_trailing_slash = dest[-1] == "/"
		dest = fix_path(dest, must_exist=False)
		if dest[-1] != "/" and has_trailing_slash:
			dest += "/"
		print(f"data will be downloaded from: '{showsrc}'")
		print(f"data will be downloaded to: '{dest}'")
		if ask and not askYesNo():
			return
		print(f"downloading data as {self.username}@{self.destination}")
		if self.server is None:
			cmd = ["rsync", "-aP", truesrc, dest]
		else:
			rsh = ["ssh", "-o", "NoHostAuthenticationForLocalhost=yes"]
			rsh = " ".join(rsh + ["-p", str(self.port)])
			cmd = ["rsync", "-aP", "--rsh", rsh, truesrc, dest]
		if dry_run:
			cmd += ["--dry-run"]
		return subprocess.run(cmd)
	
	def upload(self, src, dest, dry_run = False, ask = False):
		"""
		Upload file(s) from local storage using rsync
		:param src: The source path on the local machine
		:param dest: The destination path on the destination machine
		:param dry_run: Show what would be done without doing it?
		:param ask: Confirm before uploading?
		"""
		truedest = f"{self.username}@{self.hostname}:{dest}"
		showdest = f"{self.username}@{self.destination}:{dest}"
		has_trailing_slash = src[-1] == "/"
		src = fix_path(src, must_exist=True)
		if src[-1] != "/" and has_trailing_slash:
			src += "/"
		print(f"data will be uploaded from: '{src}'")
		print(f"data will be uploaded to: '{showdest}'")
		if ask and not askYesNo():
			return
		print(f"uploading data as {self.username}@{self.destination}")
		if self.server is None:
			cmd = ["rsync", "-aP", src, truedest]
		else:
			rsh = ["ssh", "-o", "NoHostAuthenticationForLocalhost=yes"]
			rsh = " ".join(rsh + ["-p", str(self.port)])
			cmd = ["rsync", "-aP", "--rsh", rsh, src, truedest]
		if dry_run:
			cmd += ["--dry-run"]
		return subprocess.run(cmd)
	
	def ssh(self):
		"""
		Attach an unrestricted ssh terminal session
		"""
		print(f"connecting as {self.username}@{self.destination}")
		dest = f"{self.username}@{self.hostname}"
		if self.server is None:
			cmd = ["ssh", dest]
		else:
			cmd = ["ssh", "-o", "NoHostAuthenticationForLocalhost=yes"]
			cmd += ["-p", str(self.port), dest]
		return subprocess.run(cmd)
	
	def close(self):
		"""
		Close the connection to the gateway server
		"""
		if self.process is None:
			return
		print(f"closing connection to {self.server}")
		self.process.terminate()
		self.process = None





# Experiment data manager

import os
import sys
if sys.version_info >= (3, 11):
	import tomllib
else:
	import tomli as tomllib

import shutil
from time import sleep
from dataclasses import dataclass
from dataclasses import asdict
from datetime import datetime

from .rssh import rssh
from .tools import ls
from .tools import fix_path
from .tools import dirsize
from .tools import dirfiles
from .tools import to_bytes
from .tools import format_bytes
from .tools import format_datasets
from .tools import print_datasets
from .tools import askYesNo
from .tools import grep1
from .tools import grepl

# scopes
SCOPE_PRIVATE = "Private"
SCOPE_PROTECTED = "Protected"
SCOPE_PUBLIC = "Public"
SCOPE_XFER = "Xfer"

@dataclass
class expdata:
	"""
	Experimental metadata for a scientific dataset
	"""

	name: str
	scope: str
	group: str
	title: str
	description: str
	sample_processing: str
	data_processing: str
	contact: dict
	url: dict
	date: dict
	formats: list
	keywords: list
	notes: list
	
	def __init__(self, name, entry, printwidth = 60):
		"""
		Initialize an expdata instance
		:param name: The name of the dataset
		:param entry: A dict parsed from its manifest entry
		"""
		self.name = name
		self.scope = entry["scope"]
		self.group = entry["group"]
		self.title = entry["title"]
		self.description = entry["description"]
		self.sample_processing = entry.get("sample-processing", "")
		self.data_processing = entry.get("data-processing", "")
		self.contact = entry.get("contact", [])
		self.url = entry.get("url", {})
		self.date = entry.get("date", {})
		self.formats = entry.get("formats", [])
		self.keywords = entry.get("keywords", [])
		self.notes = entry.get("notes", [])
		self.printwidth = printwidth
	
	def __str__(self):
		"""
		Return str(self)
		"""
		return self.describe(self.printwidth)
	
	def describe(self, printwidth = None):
		"""
		Return dataset description
		"""
		dataset = asdict(self)
		printed = ["scope", "group", "description",
			"sample_processing", "data_processing",
			"formats", "keywords"]
		notprinted = set(dataset.keys()).difference(printed)
		notprinted = notprinted.difference(["name", "title"])
		if len(dataset["notes"]) > 0:
			notprinted = notprinted.difference(["notes"])
		title = self.title
		if ( len(title) > 0 ):
			sl = [f" {title}: "]
		else:
			sl = [" <Untitled>: "]
		sl.append(" {")
		for field in printed:
			value = dataset[field]
			if isinstance(value, str):
				if printwidth is not None:
					if len(value) > printwidth:
						value = value[:printwidth - 4] + "..."
			else:
				value = ", ".join(value)
			sl.append(f"  {field}: {value}")
		for i, note in enumerate(dataset["notes"]):
			if printwidth is not None:
				if len(note) > printwidth:
					note = note[:printwidth - 4] + "..."
			sl.append(f"  note {i + 1}: {note}")
		more_fields = [f"'{field}'" for field in sorted(notprinted)]
		more_fields = ", ".join(more_fields)
		sl.append(f"  additional fields: {more_fields}")
		sl.append(" }")
		return "\n".join(["{"] + sl + ["}"])
	
	def search(self, pattern, context_width = 60):
		"""
		Search all metadata fields for a pattern
		:param pattern: The search pattern
		:param context_width: Width of a context window to return
		:returns: An expsearch instance or None
		"""
		search = expsearch(self, pattern, context_width=context_width)
		if len(search.hits) > 0:
			return search
		else:
			return None
	
	def has_scope(self, pattern):
		"""
		Detect if the dataset's scope matches a pattern
		:param pattern: The scope pattern
		:returns: bool
		"""
		return grep1(pattern, self.scope) is not None
	
	def has_group(self, pattern):
		"""
		Detect if the dataset's group matches a pattern
		:param pattern: The group pattern
		:returns: bool
		"""
		return grep1(pattern, self.group) is not None

@dataclass
class expsearch:
	"""
	Experimental metadata search hits
	"""

	name: str
	title: str
	scope: str
	group: str
	pattern: str
	hits: list
	
	def __init__(self, dataset, pattern, context_width = 60):
		"""
		Initialize an expsearch instance
		:param dataset: An expdata instance
		:param pattern: The search pattern
		:param context_width: Width of a context window to return
		"""
		self.name = dataset.name
		self.title = dataset.title
		self.scope = dataset.scope
		self.group = dataset.group
		self.pattern = pattern
		hits = {}
		title = grep1(pattern, dataset.title,
			context_width=context_width)
		description = grep1(pattern, dataset.description,
			context_width=context_width)
		sample_processing = grep1(pattern, dataset.sample_processing,
			context_width=context_width)
		data_processing = grep1(pattern, dataset.data_processing,
			context_width=context_width)
		if title is not None:
			hits["title"] = title
		if description is not None:
			hits["description"] = description
		if sample_processing is not None:
			hits["sample_processing"] = sample_processing
		if data_processing is not None:
			hits["data_processing"] = data_processing
		for i, contact in enumerate(dataset.contact):
			if any(grepl(pattern, contact.values())):
				hits[f"contact {i + 1}"] = contact
		for i, format_ in enumerate(dataset.formats):
			if grep1(pattern, format_) is not None:
				hits[f"format {i + 1}"] = format_
		for i, keyword in enumerate(dataset.keywords):
			if grep1(pattern, keyword) is not None:
				hits[f"keyword {i + 1}"] = keyword
		for i, note in enumerate(dataset.notes):
			if grep1(pattern, note) is not None:
				hits[f"note {i + 1}"] = note
		self.hits = hits
		self.printwidth = dataset.printwidth
	
	def __str__(self):
		"""
		Return str(self)
		"""
		title = self.title
		if ( len(title) > 0 ):
			sl = [f" {title}: "]
		else:
			sl = [" <Untitled>: "]
		sl.append(" {")
		sl.append("  _search pattern_: " + self.pattern)
		sl.append("  _no. of matches_: " + str(len(self.hits)))
		sl.append("  scope: " + str(self.scope))
		sl.append("  group: " + str(self.group))
		for field, match in self.hits.items():
			if len(match) > self.printwidth:
				match = match[:self.printwidth - 4] + "..."
			sl.append(f"  >{field}: {match}")
		sl.append(" }")
		return "\n".join(["{"] + sl + ["}"])

@dataclass
class expcache:
	"""
	File metadata for a cached dataset
	"""

	name: str
	path: str
	_atime: float
	_mtime: float
	size: int
	
	def __init__(self, name, path, atime, mtime, size):
		"""
		Initialize an expcache instance
		:param name: The name of the dataset
		:param path: The file path to the dataset
		:param atime: Last access time
		:param mtime: Last modified time
		:param size: Total size in storage
		"""
		self.name = name
		self.path = path
		self._atime = atime
		self._mtime = mtime
		self.size = size
	
	@property
	def atime(self):
		"""
		Get last accessed time
		"""
		return datetime.fromtimestamp(self._atime)
	
	@property
	def mtime(self):
		"""
		Get last modified time
		"""
		return datetime.fromtimestamp(self._mtime)
	
	def __str__(self):
		"""
		Return str(self)
		"""
		return self.describe()
	
	def describe(self, printformats = None):
		"""
		Return cached dataset description
		"""
		path = f" path: '{self.path}'"
		atime = f" atime: '{self.atime}'"
		mtime = f" mtime: '{self.mtime}'"
		size = f" size: {format_bytes(self.size)}"
		sl = [path, atime, mtime, size]
		if printformats is not None:
			files = []
			fmts = [f"{fmt}$" for fmt in printformats]
			for pattern in fmts:
				files.extend(dirfiles(self.path, pattern, recursive=True))
			files = [f" file {str(i + 1)}: '{file}'" 
				for i, file 
				in enumerate(files)]
			sl.extend(files)
			sl = [" " + s for s in sl]
			sl = [" Cached:", " {"] + sl + [" }"]
		return "\n".join(["{"] + sl + ["}"])

class expdb:
	"""
	Database manager for experimental datasets and metadata
	"""
	
	def __init__(self, username, dbpath, dbname, metaname = True,
		remote_dbhost = None, remote_dbpath = None,
		server = None, server_username = None,
		port = 8080, remote_port = 22, verbose = False,
		autoconnect = True):
		"""
		Initialize an expdb instance
		:param username: Your username on remote database host
		:param dbpath: The local database path
		:param dbname: The database name (may be None)
		:param metaname: Subdirectory containing "manifest.toml"" (optional)
		:param remote_dbhost: The remote database host
		:param remote_dbpath: The remote database path
		:param server: The gateway server hostname (optional)
		:param server_username: Your username on the gateway server (optional)
		:param port: The local port for gateway server SSH forwarding
		:param remote_port: The remote database host port
		:param verbose: Print progress messages?
		:param autoconnect: Connect on initialization?
		"""
		if remote_dbhost is not None and remote_dbpath is None:
			remote_dbpath = dbpath
		self.username = username
		self.dbpath = fix_path(dbpath, must_exist=True)
		self.dbname = dbname
		if metaname in (True, False):
			if metaname:
				metaname = dbname
			else:
				metaname = None
		self.metaname = metaname
		self.remote_dbhost = remote_dbhost
		self.remote_dbpath = remote_dbpath
		self.server = server
		self.server_username = server_username
		self.port = port
		self.remote_port = remote_port
		self.verbose = verbose
		self._manifest = None
		self._cache = None
		if autoconnect:
			self.open()
	
	def __str__(self):
		"""
		Return str(self)
		"""
		return format_datasets(self.manifest.values())
	
	def __repr__(self):
		"""
		Return repr(self)
		"""
		user = f"username='{self.username}'"
		dbpath = f"dbpath='{self.dbpath}'"
		fields = [user, dbpath]
		if self.remote_dbhost is not None:
			remote_dbhost = f"remote_dbhost='{self.remote_dbhost}'"
			fields.append(remote_dbhost)
			if self.remote_dbpath is not None:
				remote_dbpath = f"remote_dbpath='{self.remote_dbpath}'"
				fields.append(remote_dbpath)
		if self.server is not None:
			server = f"server='{self.server}'"
			if self.server_username is None:
				server_username = "server_username=None"
			else:
				server_username = f"server_username='{self.server_username}'"
			port = f"port={self.port}"
			fields.extend([server, server_username, port])
		fields = ", ".join(fields)
		return f"expdb({fields})"
		
	def __enter__(self):
		"""
		Enter context manager
		"""
		self.open_manifest()
		return self
	
	def __exit__(self, exc_type, exc_value, traceback):
		"""
		Exit context manager
		"""
		self.close()
	
	def __del__(self):
		"""
		Delete self
		"""
		self.close()
	
	def __getitem__(self, key):
		"""
		Return db[key]
		"""
		return self.get(key)
	
	@property
	def manifest(self):
		"""
		Get dataset manifest
		"""
		if self._manifest is None:
			self.open_manifest()
		return self._manifest
	
	@property
	def cache(self):
		"""
		Get local cache
		"""
		if self._cache is None:
			self.open_cache()
		return self._cache
	
	@property
	def dbdir(self):
		"""
		Resolve local cache directory
		"""
		if self.dbpath is None:
			return None
		if self.dbname is None:
			return self.dbpath
		else:
			return os.path.join(self.dbpath, self.dbname)
	
	@property
	def remote_dbdir(self):
		"""
		Resolve remote database directory
		"""
		if self.remote_dbpath is None:
			return None
		if self.dbname is None:
			return self.remote_dbpath
		else:
			return os.path.join(self.remote_dbpath, self.dbname)
	
	def get(self, key):
		"""
		Return db[key]
		:param key: A dataset name or iterable of them
		:returns: An expdata instance or dict of them
		"""
		if isinstance(key, str):
			return self.manifest.get(key)
		else:
			return {ki: self.manifest[ki] for ki in key}
	
	def isopen(self):
		"""
		Check if the database is ready
		"""
		return self._manifest is not None or self._cache is not None
	
	def open(self):
		"""
		Ready the connection to the database
		"""
		self.open_manifest()
		self.open_cache()
	
	def open_manifest(self):
		"""
		Refresh the database manifest
		"""
		if self.metaname is None:
			path = os.path.join(self.dbdir, "manifest.toml")
		else:
			path = os.path.join(self.dbdir, self.metaname, "manifest.toml")
		path = fix_path(path, must_exist=True)
		if self.verbose:
			print(f"parsing '{path}'")
		with open(path, "rb") as file:
			manifest = tomllib.load(file)
			manifest = {name: expdata(name, entry) 
				for name, entry in manifest.items()}
			self._manifest = manifest
		if self.verbose:
			print("manifest is searchable")
	
	def open_cache(self):
		"""
		Refresh the cache metadata
		"""
		if self.verbose:
			print("detecting cached datasets")
		cache = []
		cache.extend(self._get_cached_scope(SCOPE_PRIVATE))
		cache.extend(self._get_cached_scope(SCOPE_PROTECTED))
		cache.extend(self._get_cached_scope(SCOPE_PUBLIC))
		self._cache = {dataset.name: dataset for dataset in cache}
		if self.verbose:
			print(f"{len(cache)} datasets available locally")
	
	def _get_cached_dataset(self, scope, group, dataset):
		"""
		Get cached dataset metadata
		:param scope: The dataset scope
		:param group: The dataset group
		:param dataset: The dataset name
		:returns: An expcache instance
		"""
		path = os.path.join(self.dbdir, scope, group, dataset)
		path = fix_path(path, must_exist=True)
		size = dirsize(path, all_names=True)
		atime = os.path.getatime(path)
		mtime = os.path.getmtime(path)
		return expcache(name=dataset, path=path,
			atime=atime, mtime=mtime, size=size)
	
	def _get_cached_group(self, scope, group):
		"""
		Get list of cached group metadata
		:param scope: The dataset scope
		:param group: The dataset group
		:returns: A list of expcache instances
		"""
		path = os.path.join(self.dbdir, scope, group)
		path = fix_path(path, must_exist=True)
		return [self._get_cached_dataset(scope, group, dataset)
			for dataset
			in ls(path)]
	
	def _get_cached_scope(self, scope):
		"""
		Get list of cached scope metadata
		:param scope: The dataset scope
		:returns: A list of expcache instances
		"""
		path = os.path.join(self.dbdir, scope)
		groups = []
		if os.path.isdir(path):
			path = fix_path(path, must_exist=True)
			for group in ls(path):
				groups.extend(self._get_cached_group(scope, group))
		return groups
	
	def ls(self, scope = None, group = None, details = False):
		"""
		List all datasets by name
		:param scope: Filter by scope
		:param group: Filter by group
		:param details: Return names only or dataset details?
		:returns: A list of datasets
		"""
		if scope is None and group is None:
			if details:
				return list(self.manifest.values())
			else:
				return list(self.manifest.keys())
		else:
			results = []
			for name, dataset in self.manifest.items():
				ok = True
				if scope is not None:
					ok = ok and dataset.has_scope(scope)
				if group is not None:
					ok = ok and dataset.has_group(group)
				if ok:
					if details:
						results.append(dataset)
					else:
						results.append(name)
			return results
	
	def ls_cache(self, scope = None, group = None, details = False):
		"""
		List cached datasets by name
		:param scope: Filter by scope
		:param group: Filter by group
		:param details: Return names only or dataset details?
		:returns: A list of datasets
		"""
		if scope is None and group is None:
			if details:
				return list(self.cache.values())
			else:
				return list(self.cache.keys())
		else:
			results = []
			for name in self.cache.keys():
				ok = True
				dataset = self.get(name)
				if dataset is None:
					continue
				if scope is not None:
					ok = ok and dataset.has_scope(scope)
				if group is not None:
					ok = ok and dataset.has_group(group)
				if ok:
					if details:
						results.append(self.cache.get(name))
					else:
						results.append(name)
			return results
	
	def search(self, pattern = None, scope = None, group = None):
		"""
		Search all dataset metadata for a pattern
		:param pattern: The pattern to find
		:param scope: Filter by scope
		:param group: Filter by group
		:returns: A list of search hits
		"""
		hits = []
		for name, dataset in self.manifest.items():
			ok = True
			if scope is not None:
				ok = ok and dataset.has_scope(scope)
			if group is not None:
				ok = ok and dataset.has_group(group)
			if ok and pattern is not None:
				result = dataset.search(pattern)
				if result is not None:
					hits.append(result)
		return hits
	
	def search_cache(self, pattern = None, scope = None, group = None):
		"""
		Search cached dataset metadata for a pattern
		:param pattern: The pattern to find
		:param scope: Filter by scope
		:param group: Filter by group
		:returns: A list of search hits
		"""
		hits = []
		for name in self.cache.keys():
			ok = True
			dataset = self.get(name)
			if dataset is None:
				continue
			if scope is not None:
				ok = ok and dataset.has_scope(scope)
			if group is not None:
				ok = ok and dataset.has_group(group)
			if ok and pattern is not None:
				result = dataset.search(pattern)
				if result is not None:
					hits.append(result)
		return hits
	
	def prune_cache(self, limit, units, scope = None, group = None,
		strategy = "lru", dry_run = False, ask = False):
		"""
		List cached datasets by name
		:param limit: Maximum size of the cache
		:param units: Size units (B, KB, MB, GB, TB, etc.)
		:param strategy: One of "lru", "mru", "big", "small"
		:param dry_run: Show what would be done without doing it?
		:param ask: Confirm before deleting?
		"""
		limit = to_bytes(limit, units)
		cache = self.ls_cache(scope, group, details=True)
		strategy = strategy.casefold()
		if strategy == "lru".casefold():
			cache.sort(key=lambda x: x.atime)
		elif strategy == "mru".casefold():
			cache.sort(key=lambda x: x.atime)
			cache.reverse()
		elif strategy == "big".casefold():
			cache.sort(key=lambda x: x.size)
			cache.reverse()
		elif strategy == "small".casefold():
			cache.sort(key=lambda x: x.size)
		else:
			raise ValueError(f"invalid strategy: {strategy}")
		sizes = [x.size for x in cache]
		totsize = sum(sizes)
		delsize = []
		for i in range(len(sizes)):
			delsize.append(totsize - sum(sizes[:i]))
		target = 0
		while delsize[target] > limit:
			target += 1
		newsize = totsize - sum(sizes[:target])
		print(f"the local cache is currently {format_bytes(totsize)}")
		print(f"using strategy: {strategy}")
		print("the following datasets will be deleted from the cache:")
		print_datasets(cache[:target], names_only=True)
		print(f"~= {format_bytes(sum(sizes[:target]))} will be freed")
		if not target:
			return
		if ask and not askYesNo():
			return
		if not dry_run:
			for x in cache[:target]:
				print(f"deleting '{x.path}'")
				try:
					shutil.rmtree(fix_path(x.path, must_exist=True))
				except Exception:
					print(f"failed to delete '{x.path}'")
			print(f"the local cache is now {format_bytes(newsize)}")
		return
	
	def sync(self, name, force = False, dry_run = False, ask = False):
		"""
		Sync a dataset to local storage
		:param name: The name of the dataset
		:param force: Should the dataset be re-synced if already cached?
		:param dry_run: Show what would be done without doing it?
		:param ask: Confirm before downloading?
		"""
		if name not in self.manifest:
			raise KeyError(f"no such dataset: '{name}'")
		if name in self.cache and not force:
			print("dataset is already cached; use force=True to re-sync")
			return
		if self.remote_dbhost is None:
			raise IOError("remote host is None")
		if self.remote_dbpath is None:
			raise IOError("remote path is None")
		dataset = self.get(name)
		scope = dataset.scope
		group = dataset.group
		path = os.path.join(self.dbdir, scope, group)
		path = fix_path(path, must_exist=False)
		if not os.path.isdir(path):
			os.makedirs(path)
		src = os.path.join(self.remote_dbdir, scope, group, name)
		src = src + "/"
		dest = os.path.join(self.dbdir, scope, group, name)
		dest = dest + "/"
		try:
			con = rssh(self.username,
				destination=self.remote_dbhost,
				server=self.server,
				server_username=self.server_username,
				port=self.port,
				destination_port=self.remote_port)
			sleep(1) # allow time to connect
			con.download(src, dest, dry_run=dry_run, ask=ask)
			if os.path.isdir(dest):
				print("sync complete; refreshing cache metadata")
				self.open_cache()
		except Exception:
			print("a problem occured during syncing")
		finally:
			con.close()
	
	def submit(self, path, force = False, dry_run = False, ask = False):
		"""
		Submit a dataset to the database maintainer(s)
		:param path: The path to the data directory
		:param force: Should the dataset be re-submitted if already tracked?
		:param dry_run: Show what would be done without doing it?
		:param ask: Confirm before uploading?
		"""
		path = fix_path(path, must_exist=False)
		name = os.path.basename(path)
		if name in self.manifest and not force:
			print("dataset is already tracked; use force=True to re-submit")
			return
		if self.remote_dbhost is None:
			raise IOError("remote host is None")
		if self.remote_dbpath is None:
			raise IOError("remote path is None")
		if not os.path.isdir(path):
			raise NotADirectoryError(f"path must be a directory: {path}")
		if path[-1] != "/":
			src = path + "/"
		else:
			src = path
		dest = os.path.join(self.remote_dbdir, SCOPE_XFER, name)
		dest = dest + "/"
		try:
			con = rssh(self.username,
				destination=self.remote_dbhost,
				server=self.server,
				server_username=self.server_username,
				port=self.port,
				destination_port=self.remote_port)
			sleep(1) # allow time to connect
			con.upload(src, dest, dry_run=dry_run, ask=ask)
			print("submission complete")
		except Exception:
			print("a problem occured during submission")
		finally:
			con.close()
	
	def status(self, scope = None, group = None, details = False):
		"""
		Get status of cache versus manifest
		:param scope: Filter by scope
		:param group: Filter by group
		:param details: Return names only or dataset details?
		:returns: A tuple like (synced, remoteonly, localonly)
		"""
		remote = set(self.manifest.keys())
		local = set(self.cache.keys())
		synced = {name: dataset 
			for name, dataset 
			in self.cache.items()
			if name in remote.intersection(local)}
		remoteonly = {name: dataset 
			for name, dataset 
			in self.manifest.items()
			if name in remote.difference(local)}
		localonly = {name: dataset 
			for name, dataset 
			in self.cache.items()
			if name in local.difference(remote)}
		if scope is None and group is None:
			if details:
				synced_list = list(synced.values())
				remoteonly_list = list(remoteonly.values())
				localonly_list = list(localonly.values())
			else:
				synced_list = list(synced.keys())
				remoteonly_list = list(remoteonly.keys())
				localonly_list = list(localonly.keys())
		else:
			synced_list = []
			remoteonly_list = []
			localonly_list = []
			for name in synced.keys():
				ok = True
				dataset = self.manifest.get(name)
				if dataset is None:
					continue
				if scope is not None:
					ok = ok and dataset.has_scope(scope)
				if group is not None:
					ok = ok and dataset.has_group(group)
				if ok:
					if details:
						synced_list.append(synced.get(name))
					else:
						synced_list.append(name)
			for name in remoteonly.keys():
				ok = True
				dataset = self.manifest.get(name)
				if dataset is None:
					continue
				if scope is not None:
					ok = ok and dataset.has_scope(scope)
				if group is not None:
					ok = ok and dataset.has_group(group)
				if ok:
					if details:
						remoteonly_list.append(remoteonly.get(name))
					else:
						remoteonly_list.append(name)
			if details:
				localonly_list = list(localonly.values())
			else:
				localonly_list = list(localonly.keys())
		return synced_list, remoteonly_list, localonly_list
	
	def close(self):
		"""
		Remove the connection to the database
		"""
		self._manifest = None
		self._cache = None
